# automatically generated by the FlatBuffers compiler, do not modify

# namespace: NetEncoding

class TypedArray(object):
    NONE = 0
    Float32Array = 1
    Int32Array = 2
    Uint32Array = 3
    Float64Array = 4
    JSONEncodedArray = 5

