<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <title>cell&times;gene</title>
    <style>
      html,
      body,
      p,
      h1,
      h2,
      h3,
      h4,
      h5,
      h6,
      span,
      button,
      input,
      label,
      text,
      div {
        font-family: "Roboto Condensed", "Helvetica Neue", "Helvetica", "Arial",
          sans-serif;
        font-size: 14px;
      }

      body {
        margin: 0;
        padding: 0;
      }

      * {
        box-sizing: border-box;
      }
    </style>
  <link rel="shortcut icon" href="static/assets/favicon.ico"><link rel="icon" type="image/png" sizes="16x16" href="static/assets/favicon-16x16.png"><link rel="icon" type="image/png" sizes="32x32" href="static/assets/favicon-32x32.png"><link rel="icon" type="image/png" sizes="48x48" href="static/assets/favicon-48x48.png"><script defer src="static/main-2a269ec44f688b5ec34d.js"></script><script defer src="static/obsolete-4de0170af9af7dbde9f69250fc4e7270.js" async></script><link href="static/main-1fed8eab18c965a39492.css" rel="stylesheet"></head>
  <body>
    <script type="text/javascript">
      window.CELLXGENE = {};
      window.CELLXGENE.API = {
        prefix: `${location.origin}${location.pathname}api/`,
        version: "v0.2/",
      };
    </script>
    <noscript
      >If you're seeing this message, that means
      <strong>JavaScript has been disabled on your browser</strong>, please
      <strong>enable JS</strong> to make this app work.
    </noscript>

    <div id="root"></div>
    {% for script in SCRIPTS -%}
    <script type="text/javascript"
      {{ ('integrity="%s"' % script.integrity) | safe if script.integrity }}
      {{ ('crossorigin="%s"' % script.crossorigin) | safe if script.crossorigin }}
      src="{{script.src | safe}}"
    ></script>
    {%- endfor %}

    {% for ils in INLINE_SCRIPTS %}
    <script type="text/javascript" >{% include ils %}</script>
    {% endfor %}
  </body>
</html>
